@@ -1,9 +1,9 @@
-/* min12xxw.c		driver for minolta pagepro 1[23]xxW printers
+/* min12xxw.c		driver for minolta pagepro 1[234]xxW printers
  *
  * It converts pages in pbmraw format read from stdin to the printer
- * language used by Minolta PagePro 1[23]00W. The output goes to stdout.
+ * language used by Minolta PagePro 1[234]00W. The output goes to stdout.
  *
- * Copyright (C) 2004  Manuel Tobias Schiller <mala@hinterbergen.de>
+ * Copyright (C) 2004-2006 Manuel Tobias Schiller <mala@hinterbergen.de>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
+ * Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 /* Note: This program was written without any documentation from the
  * manufacturer. The description of the printer language may be
@@ -38,7 +38,7 @@
 /* see the file CHANGELOG for a list of changes between revisions.
  */
 
-#include <config.h>
+#include "config.h"
 
 #define _GNU_SOURCE /* we might need this because we use getline */
 #include <assert.h>
@@ -46,8 +46,8 @@
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
-#include <unistd.h> /* usleep */
-#include <sys/stat.h> /* fstat */
+#include <unistd.h>	/* usleep */
+#include <sys/stat.h>	/* fstat */
 #include <getopt.h>
 
 /* include the inttypes.h for uint8_t and similar types */
@@ -72,17 +72,17 @@
 #define MAX(x, y) (((x) > (y))?(x):(y))
 #endif
 
-static char *version = "version " PACKAGE_VERSION " ($Id: min12xxw.c 48 2005-03-28 16:46:44Z mala $)";
+/* we wish to know what version we are ;) */
+static char *versionstr = "version " PACKAGE_VERSION " ($Id: min12xxw.c 115 2005-12-31 13:50:48Z mala $)";
 
 /* defaults used for the output */
 static int ptype = 0x00; /* normal paper */
 static int pformat = 0x04; /* a4 */
-static int res = 1; /* 600 dpi */
+static int res = 0x0001; /* 600 dpi, horizontal resolution not doubled */
 static int tray = 0xff; /* auto select paper tray */
 static int nomargins = 0; /* if flag is set, margins are not enforced */
 static int ecomode = 0; /* if set, toner-saving mode is enabled */
 static int model = 0x81; /* 12xxW series printers are the default */
-static int hresdbl = 0; /* don't double horizontal resolution */
 static char *device = "/dev/lp0"; /* default device used for queries */
 
 /* try to exit gracefully in case of fatal errors */
@@ -209,7 +209,10 @@ static void do_enreg(FILE *out)
 {
 	uint8_t buf[3];
 
-	buf[0] = ((model == 0x83)?(0x1c):(0x78));
+	/* FIXME: until we know better, we'll just treat the 1400W as a
+	 * better 13xxW here - testers report that this seems to work
+	 * well */
+	buf[0] = (((model == 0x83) || (model == 0x86))?(0x1c):(0x78));
 	buf[1] = 0;
 	buf[2] = 0x04;
 	do_cmd(out, 0x6a, 3, buf);
@@ -340,13 +343,13 @@ static void send_start_job(FILE *out)
 
 	do_start(out); /* start-of-printer-commands command */
 
-	cmdbuf[0] = (uint8_t) res; /* set resolution and paper type */
+	cmdbuf[0] = (uint8_t) (res & 0xff); /* set resolution and paper type */
+	/* check if we are to use doubled horizontal resolution */
+	cmdbuf[1] = (uint8_t) (res >> 8);
 	cmdbuf[3] = (uint8_t) ptype;
 	cmdbuf[4] = 0x04;
-	/* 13xxW series models might expect an 0x04 in cmdbuf[6] as well */
-	if (model == 0x83) cmdbuf[6] = 0x04;
-	/* check if we are to use doubled horizontal resolution */
-	if (hresdbl) cmdbuf[1] = 0x01;
+	/* 1[34]xxW series models might expect an 0x04 in cmdbuf[6] as well */
+	if ((model == 0x83) || (model == 0x86)) cmdbuf[6] = 0x04;
 	do_cmd(out, 0x50, 8, cmdbuf); /* issue it */
 }
 
@@ -372,7 +375,7 @@ static void send_new_page(FILE *out, uint32_t x, uint32_t y)
 	*p++ = y & 0xff; *p++ = (y >> 8) & 0xff;
 	*p++ = 0x08; p++; *p++ = 0x08; p++;
 	*p++ = tray; *p++ = pformat;
-	if (res == 0) {
+	if ((res & 0xff) == 0) {
 		/* apparently, 300 dpi needs special flags set here */
 		cmdbuf[20] = 0xc0;
 	}
@@ -545,7 +548,7 @@ static void dopage(FILE *in, FILE *out, uint32_t x, uint32_t y)
 	uint32_t yy, yc; /* counters */
 	/* this controls how many pixels are skipped on each side of a
 	 * sheet of paper to keep the margins that the printer expects */
-	int skip = ((nomargins == 0)?skiptbl[res]:0);
+	int skip = ((nomargins == 0)?skiptbl[res & 0xff]:0);
 	int i; /* counter */
 	int ecofl = 0; /* flag used for toner-saving mode */
 
@@ -610,7 +613,7 @@ static void dojob(FILE *in, FILE *out)
 	char *line = NULL; /* buffer for reading whole lines from the input */
 	size_t llen = 0; /* length of line buffer */
 	/* how much to skip because of margins */
-	int skip = ((nomargins == 0)?(16 * skiptbl[res]):0);
+	int skip = ((nomargins == 0)?(16 * skiptbl[res & 0xff]):0);
 
 	/* send the printer a start-of-job/set-resolution-and-papertype
 	 * sequence */
@@ -622,19 +625,40 @@ static void dojob(FILE *in, FILE *out)
 		if (getline(&line, &llen, in) == -1) break;
 		/* check for valid pbmraw signature */
 		if ((line[0] != 'P') || (line[1] != '4'))
-			fatal("min12xxw: input is not valid pbmraw");
+			fatal("min12xxw: input is not valid pbmraw "
+				"(no valid signature)");
 		/* read in comment lines and the line containing the
 		 * resolution */
 		do {
 			if (getline(&line, &llen, in) == -1)
-				fatal("min12xxw: input is not valid pbmraw");
+				fatal("min12xxw: input is not valid pbmraw "
+					"(premature end of file)");
 		} while (line[0] == '#');
 		/* parse resolution */
-		if (sscanf(line, "%lu %lu", (unsigned long *) &x, \
-				(unsigned long *) &y) != 2)
-			fatal("min12xxw: input is not valid pbmraw");
+		{
+			/* this ugly hack is needed to make things work
+			 * on 64 bit platforms like sparc64 or amd64 */
+			unsigned long lx, ly;
+			if (sscanf(line, "%lu %lu", &lx, &ly) != 2)
+				fatal("min12xxw: input is not valid pbmraw "
+					"(ill formatted bitmap dimensions)");
+			x = lx; y = ly;
+		}
 		/* scanlines are byte-aligned - adjust x accordingly */
 		if (x & 0x7) x = 8 + (x & ~0x7);
+		/* check if the page dimensions are so small that we need
+		 * switch of the margins
+		 * many thanks go to Ben Cooper who alerted me to the
+		 * problem */
+		if ((((y - skip) / 8) <= skip) || (x <= (2 * skip))) {
+			if (!nomargins)
+				fprintf(stderr, "min12xxw: page dimensions"
+					" are so small that I won't "
+					"enforce page margins for this and"
+					"all subsequent pages!\n");
+			nomargins = 1;
+			skip = 0;
+		}
 
 		/* tell the printer to start a new page - note the page
 		 * dimensions are reduced because we need to take the
@@ -655,44 +679,9 @@ static void dojob(FILE *in, FILE *out)
 /**********************************************************************
  * option handling
  **********************************************************************/
-/* this procedure gives help */
-static void help()
-{
-	fprintf(stderr, "min12xxw - a pbmraw to Milolta PagePro 1[23]xx W " \
-			"filter\n\noptions: (defaults are marked with an" \
-			" asterisk)\n" \
-			"\t-h\t--help\t\t\tthis help\n" \
-			"\t-v\t--version\t\tshow version number\n" \
-			"\t-d dev\t--device dev\t\tset device to use for " \
-			"queries\n" \
-			"\t-s\t--status\t\tquery printer status\n" \
-			"\t-e\t--ecomode\t\teconomic (toner saving) mode\n" \
-			"\t-n\t--nomargins\t\tdisable enforcement of " \
-			"margins\n" \
-			"\t-m mod\t--model mod\t\tset the printer model to " \
-			"produce output for\n" \
-			"\t-r res\t--res res\t\tset resolution\n" \
-			"\t-t tray\t--tray tray\t\tset paper tray\n" \
-			"\t-p type\t--papertype type\tset paper type" \
-			"\n\t-f fmt\t--paperformat fmt\tset paper format\n" \
-			"\navailable printer models: 1200W*, 1250W, 1300W, " \
-			"1350W\n" \
-			"\navailable resolutions in dpi: 300, 600*, 1200\n" \
-			"\navailable paper trays: auto*, tray1, tray2, " \
-			"manual\n\navailable paper types: normal*, thick, "\
-			"envelope, postcard\n" \
-			"\navailable paper formats: " \
-			"a4*, b5, a5, jpost, corpost,\njisy6, jisy0, "\
-			"chinese16k, chinese32k, legal, glegal,\n" \
-			"letter, gletter, executive, halfletter, " \
-			"envmonarch,\nenv10, envdl, envc5, envc6," \
-			"envb5, choukei3gou, choukei5gou,\ncustom, " \
-			"envb6, folio, jisy1, jisy2, quadpost\n\n");
-	exit(1);
-}
-
 /* the next few tables contain the mappings from human-readable strings
- * to id numbers the printer understands */
+ * to id numbers the printer understands (or our routines, for that
+ * matter) */
 struct map {
 	char *str;
 	int id;
@@ -719,8 +708,91 @@ static struct map pformats[] = {
 };
 static struct map models[] = {
 	{ "1200W", 0x81 }, { "1250W", 0x81 },
-	{ "1300W", 0x83 }, { "1350W", 0x83 }, { NULL, -1 }
+	{ "1300W", 0x83 }, { "1350W", 0x83 },
+	{ "1400W", 0x86 }, { NULL, -1 }
 };
+static struct map resolutions[] = {
+	{  "300", 0x0000 }, {   "300x300", 0x0000 },
+	{  "600", 0x0001 }, {   "600x600", 0x0001 },
+	{ "1200", 0x0002 }, { "1200x1200", 0x0002 },
+	{ "1200x600", 0x0101 }, { NULL, -1 } 
+};
+
+/* this procedure is used by help() below to print available somethings
+ * in a nice and painless way from the maps above
+ * see help() below and everything will become clearer */
+static void printav(char *msg, struct map *m, int defid)
+{
+	int i, n;
+	
+	n = printf(msg);
+	/* print all available somethings from our map */
+	for (i = 0; m[i].str != NULL; i++) {
+		/* in case we had to start a new line below, we need to
+		 * print a tab here */
+		if (n == 0) {
+			printf("\t");
+			n = 8;
+		}
+		/* print the next something */
+		n += printf("%s", m[i].str);
+		/* is it the default? if so, print "*" */
+		if (m[i].id == defid)
+			n += printf("*");
+		/* print ", " unless it's the last something */
+		if (m[i + 1].str != NULL)
+			n += printf(", ");
+		/* check if we have to start a new line */
+		if (n > 72) {
+			printf("\n");
+			n = 0;
+		}
+	}
+	/* unless the last thing we did was to start a new line, we append
+	 * a newline here */
+	if (n != 0) printf("\n");
+}
+
+/* this procedure gives help */
+static void help()
+{
+	printf("min12xxw - a pbmraw to Minolta PagePro 1[234]xx W "
+		"filter\n\noptions: (defaults are marked with an"
+		" asterisk)\n"
+		"\t-h\t--help\t\t\tthis help\n"
+		"\t-v\t--version\t\tshow version number\n"
+		"\t-d dev\t--device dev\t\tset device to use for "
+			"queries (%s by default)\n"
+		"\t-s\t--status\t\tquery printer status\n"
+		"\t-e\t--ecomode\t\teconomic (toner saving) mode\n"
+		"\t-n\t--nomargins\t\tdisable enforcement of margins\n"
+		"\t-m mod\t--model mod\t\tset the printer model to "
+			"produce output for\n"
+		"\t-r res\t--res res\t\tset resolution\n"
+		"\t-t tray\t--tray tray\t\tset paper tray\n"
+		"\t-p type\t--papertype type\tset paper type"
+		"\n\t-f fmt\t--paperformat fmt\tset paper format\n"
+		"\n", device);
+	/* print the available somethings */
+	printav("available models: ", models, model);
+	printav("available resolutions: ", resolutions, res);
+	printav("available paper trays: ", trays, tray);
+	printav("available paper types: ", ptypes, ptype);
+	printav("available paper formats: ", pformats, pformat);
+	printf("\n");
+	exit(0);
+}
+
+/* this procedure prints out version information */
+static void version()
+{
+	printf("min12xxw: %s\n", versionstr);
+	printf("\nCopyright (C) 2004-2006 Manuel Tobias Schiller\n"
+		"This is free software; see the source for copying "
+		"conditions.  There is NO\nwarranty; not even for "
+		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
+	exit(0);
+}
 
 /* convert string to id number */
 static int getid(struct map *m, char *thing, char *str, int deflt)
@@ -733,38 +805,6 @@ static int getid(struct map *m, char *thing, char *str, int deflt)
 	return deflt; /* return the default */
 }
 
-/* convert resolution to id number */
-static int getres(char *res)
-{
-	int v, h, i;
-
-	if ((i = sscanf(res, "%dx%d", &h, &v)) >= 1)	{
-		if ((i == 2) && (h != v)) {
-			if ((2 * v != h) || (v < 600)) {
-				/* horizontal must be twice the vertical
-				 * resolution; vertical resolution less than
-				 * 600 dpi does not work */
-				fprintf(stderr, \
-					"min12xxw: unknown resolution: %s\n",
-					res);
-				return 1;
-			}
-			hresdbl = 1; /* double horizontal resolution */
-			h = v; /* we expect vertical resolutions here */
-		}
-		switch (h) {
-		case 300:
-			return 0;
-		case 600:
-			return 1;
-		case 1200:
-			return 2;
-		}
-	}
-	fprintf(stderr, "min12xxw: unknown resolution: %s\n", res);
-	return 1;
-}
-
 /* parse the options given to the filter */
 static void parseopts(int argc, char **argv)
 {
@@ -785,6 +825,9 @@ static void parseopts(int argc, char **argv)
 	};
 	int optidx = 0, dostate = 0;
 	int c;
+	int lres = res, lptype = ptype, lpformat = pformat, lmodel = model;
+	int ltray = tray, lnomargins = nomargins, lecomode = ecomode;
+	char *ldevice = device;
 
 	opterr = 1; /* let getopt_long do the error reporting */
 	while (42) { /* this line is for all hitch-hikers out there */
@@ -794,34 +837,33 @@ static void parseopts(int argc, char **argv)
 		case 'h':
 			help();
 		case 'v':
-			printf("min12xxw: %s\n", version);
-			exit(0);
+			version();
 		case 'r':
-			res = getres(optarg);
+			lres = getid(resolutions, "resolution", optarg, 0x0001);
 			break;
 		case 'p':
-			ptype = getid(ptypes, "paper type", optarg, 0);
+			lptype = getid(ptypes, "paper type", optarg, 0);
 			break;
 		case 'f':
-			pformat = getid(pformats, "paper format", optarg, 4);
+			lpformat = getid(pformats, "paper format", optarg, 4);
 			break;
 		case 'm':
-			model = getid(models, "printer model", optarg, 0x81);
+			lmodel = getid(models, "printer model", optarg, 0x81);
 			break;
 		case 't':
-			tray = getid(trays, "tray", optarg, 0xff);
+			ltray = getid(trays, "tray", optarg, 0xff);
 			break;
 		case 'n':
-			nomargins = 1;
+			lnomargins = 1;
 			break;
 		case 'd':
-			device = optarg;
+			ldevice = optarg;
 			break;
 		case 's':
 			dostate = 1;
 			break;
 		case 'e':
-			ecomode = 1;
+			lecomode = 1;
 			break;
 		default:
 			/* getopt_long told the user about the unknown option
@@ -829,6 +871,12 @@ static void parseopts(int argc, char **argv)
 			exit(1);
 		}
 	}
+	/* copy the local copies of options to the global ones
+	 * we need to do this copying to make sure that help sees the
+	 * global defaults, not what the users sets with options */
+	res = lres; ptype = lptype; pformat = lpformat; model = lmodel;
+	tray = ltray; nomargins = lnomargins; device = ldevice;
+	ecomode = lecomode;
 	if (dostate) {
 		/* process printer queries right away and exit */
 		FILE *d;
@@ -856,6 +904,7 @@ static void modeladj(char *str)
 	if (strcmp(str, "min1250w") == 0) model = 0x81;
 	if (strcmp(str, "min1300w") == 0) model = 0x83;
 	if (strcmp(str, "min1350w") == 0) model = 0x83;
+	if (strcmp(str, "min1400w") == 0) model = 0x86;
 }
 
 /**********************************************************************
